from collections.abc import Callable
from typing import Any, Optional

import torch

from elasticai.creator.base_modules.arithmetics import Arithmetics
from elasticai.creator.base_modules.linear import Linear


class LSTMCell(torch.nn.Module):
    def __init__(
        self,
        input_size: int,
        hidden_size: int,
        bias: bool,
        arithmetics: Arithmetics,
        sigmoid_factory: Callable[[], torch.nn.Module],
        tanh_factory: Callable[[], torch.nn.Module],
        device: Any = None,
    ) -> None:
        super().__init__()
        self.input_size = input_size
        self.hidden_size = hidden_size
        self.bias = bias
        self.ops = arithmetics

        self.linear_ih = Linear(
            in_features=input_size,
            out_features=hidden_size * 4,
            bias=bias,
            arithmetics=arithmetics,
            device=device,
        )
        self.linear_hh = Linear(
            in_features=hidden_size,
            out_features=hidden_size * 4,
            bias=bias,
            arithmetics=arithmetics,
            device=device,
        )
        self.sigmoid = sigmoid_factory()
        self.tanh = tanh_factory()

    def _initialize_previous_state(
        self, x: torch.Tensor, state: Optional[tuple[torch.Tensor, torch.Tensor]]
    ) -> tuple[torch.Tensor, torch.Tensor]:
        if state is None:
            zeros = torch.zeros(*(*x.shape[:-1], self.hidden_size), dtype=x.dtype)
            return torch.clone(zeros), torch.clone(zeros)
        return state

    def forward(
        self, x: torch.Tensor, state: Optional[tuple[torch.Tensor, torch.Tensor]] = None
    ) -> tuple[torch.Tensor, torch.Tensor]:
        h_prev, c_prev = self._initialize_previous_state(x, state)

        pred_ii, pred_if, pred_ig, pred_io = torch.split(
            self.linear_ih(x), self.hidden_size, dim=-1
        )
        pred_hi, pred_hf, pred_hg, pred_ho = torch.split(
            self.linear_hh(h_prev), self.hidden_size, dim=-1
        )

        i = self.sigmoid(self.ops.add(pred_ii, pred_hi))
        f = self.sigmoid(self.ops.add(pred_if, pred_hf))
        g = self.tanh(self.ops.add(pred_ig, pred_hg))
        o = self.sigmoid(self.ops.add(pred_io, pred_ho))

        c = self.ops.add(self.ops.mul(f, c_prev), self.ops.mul(i, g))
        h = self.ops.mul(o, self.tanh(c))

        return h, c



import torch

from elasticai.creator.base_modules.arithmetics import Arithmetics
from elasticai.creator.base_modules.autograd_functions.fixed_point_quantization import (
    FixedPointDequantFunction,
    FixedPointQuantFunction,
)
from elasticai.creator.nn.two_complement_fixed_point_config import FixedPointConfig


class FixedPointArithmetics(Arithmetics):
    def __init__(self, config: FixedPointConfig) -> None:
        self.config = config

    def quantize(self, a: torch.Tensor) -> torch.Tensor:
        return self.round(self.clamp(a))

    def clamp(self, a: torch.Tensor) -> torch.Tensor:
        return torch.clamp(
            a, min=self.config.minimum_as_rational, max=self.config.maximum_as_rational
        )

    def round(self, a: torch.Tensor) -> torch.Tensor:
        def float_to_int(x: torch.Tensor) -> torch.Tensor:
            return FixedPointQuantFunction.apply(x, self.config)

        def int_to_fixed_point(x: torch.Tensor) -> torch.Tensor:
            return FixedPointDequantFunction.apply(x, self.config)

        return int_to_fixed_point(float_to_int(a))

    def add(self, a: torch.Tensor, b: torch.Tensor) -> torch.Tensor:
        return self.clamp(a + b)

    def sum(self, tensor: torch.Tensor, *tensors: torch.Tensor) -> torch.Tensor:
        summed = tensor
        for t in tensors:
            summed += t
        return self.clamp(summed)

    def mul(self, a: torch.Tensor, b: torch.Tensor) -> torch.Tensor:
        return self.round(self.clamp(a * b))

    def matmul(self, a: torch.Tensor, b: torch.Tensor) -> torch.Tensor:
        return self.round(self.clamp(torch.matmul(a, b)))



from abc import ABC, abstractmethod
from collections.abc import Iterator
from itertools import chain

from elasticai.creator.hdl.design_base.signal import Signal
from elasticai.creator.hdl.savable import Path, Savable


class Port:
    def __init__(self, incoming: list[Signal], outgoing: list[Signal]):
        self._incoming = set(incoming)
        self._outgoing = set(outgoing)

    @property
    def incoming(self) -> list[Signal]:
        return list(self._incoming)

    @property
    def outgoing(self) -> list[Signal]:
        return list(self._outgoing)

    @property
    def signals(self) -> list[Signal]:
        return self.outgoing + self.incoming

    @property
    def signal_names(self) -> list[str]:
        return [s.name for s in self.signals]

    def __getitem__(self, item: str) -> Signal:
        for signal in self.signals:
            if signal.name == item:
                return signal
        raise AttributeError(f"Port has no signal {item}")

    def __contains__(self, item: str | Signal) -> bool:
        if isinstance(item, str):
            return item in self.signal_names
        else:
            return item in self.signals

    def __iter__(self) -> Iterator[Signal]:
        yield from chain(self._incoming, self._outgoing)


class Design(Savable, ABC):
    def __init__(self, name: str):
        self.name = name

    @property
    @abstractmethod
    def port(self) -> Port:
        ...

    @abstractmethod
    def save_to(self, destination: Path) -> None:
        ...



from math import ceil, floor

from .precomputed_scalar_function import PrecomputedMonotonouslyIncreasingScalarFunction


class HardSigmoid(PrecomputedMonotonouslyIncreasingScalarFunction):
    def __init__(self, width: int, lower_bound_for_zero: int, upper_bound_for_one: int):
        self.lower_bound_for_zero = lower_bound_for_zero
        self.upper_bound_for_one = upper_bound_for_one
        self.slope = 1 / (self.upper_bound_for_one - self.lower_bound_for_zero)
        self.zero_point = -(self.slope * self.lower_bound_for_zero)
        super().__init__(
            name="hard_sigmoid",
            width=width,
            function=self._function,
            inputs=list(range(lower_bound_for_zero, upper_bound_for_one)),
        )

    @staticmethod
    def _rounding_half_to_even(x: float) -> int:
        distance_to_ceil = ceil(x) - x
        distance_to_floor = x - floor(x)
        if distance_to_floor < distance_to_ceil:
            return floor(x)
        if distance_to_ceil < distance_to_floor:
            return ceil(x)
        else:
            if ceil(x) % 2 == 0:
                return ceil(x)
            else:
                return floor(x)

    def _function(self, x: int) -> int:
        if x < self.lower_bound_for_zero:
            return 0
        if x > self.upper_bound_for_one:
            return 1
        return self._rounding_half_to_even(x * self.slope + self.zero_point)



from abc import abstractmethod
from typing import Iterable, Protocol

from elasticai.creator.hdl.code_generation.template import Template


class File(Protocol):
    @abstractmethod
    def write(self, template: Template) -> None:
        ...


class Path(Protocol):
    @abstractmethod
    def as_file(self, suffix: str) -> File:
        ...

    @abstractmethod
    def create_subpath(self, subpath_name: str) -> "Path":
        ...


class Savable(Protocol):
    @abstractmethod
    def save_to(self, destination: Path) -> None:
        ...



from elasticai.creator.hdl.code_generation.code_generation import (
    calculate_address_width,
)
from elasticai.creator.hdl.design_base import std_signals as _signals
from elasticai.creator.hdl.design_base.design import Port


def create_port_for_base_design(x_width: int, y_width: int):
    return Port(
        incoming=[
            _signals.enable(),
            _signals.clock(),
            _signals.x(x_width),
        ],
        outgoing=[_signals.y(y_width)],
    )


def create_port_for_buffered_design(
    x_width: int, y_width: int, x_count: int, y_count: int
) -> Port:
    in_signals = [
        _signals.enable(),
        _signals.clock(),
        _signals.x(x_width),
        _signals.y_address(calculate_address_width(y_count)),
    ]
    out_signals = [
        _signals.done(),
        _signals.y(y_width),
        _signals.x_address(calculate_address_width(x_count)),
    ]
    return Port(incoming=in_signals, outgoing=out_signals)



from elasticai.creator.base_modules.relu import ReLU
from elasticai.creator.hdl.design_base.design import Design
from elasticai.creator.hdl.translatable import Translatable

from .design import FPReLU as FPReLUDesign


class FPReLU(Translatable, ReLU):
    def __init__(self, total_bits: int, use_clock: bool = False) -> None:
        super().__init__()
        self._total_bits = total_bits
        self._use_clock = use_clock

    def translate(self, name: str) -> Design:
        return FPReLUDesign(
            name=name,
            data_width=self._total_bits,
            use_clock=self._use_clock,
        )



from typing import cast

import torch

from elasticai.creator.base_modules.tanh import Tanh
from elasticai.creator.hdl.design_base.design import Design
from elasticai.creator.hdl.translatable import Translatable
from elasticai.creator.hdl.vhdl.designs.monotonously_increasing_precomputed_scalar_function.precomputed_scalar_function import (
    PrecomputedMonotonouslyIncreasingScalarFunction,
)
from elasticai.creator.nn.fixed_point_arithmetics import FixedPointArithmetics
from elasticai.creator.nn.two_complement_fixed_point_config import FixedPointConfig


class FPTanh(Tanh, Translatable):
    def __init__(
        self,
        total_bits: int,
        frac_bits: int,
        num_steps: int,
        sampling_intervall: tuple[float, float] = (-5, 5),
    ) -> None:
        super().__init__(
            arithmetics=FixedPointArithmetics(
                config=FixedPointConfig(total_bits=total_bits, frac_bits=frac_bits)
            ),
            num_steps=num_steps,
            sampling_intervall=sampling_intervall,
        )
        self._config = cast(FixedPointArithmetics, self._arithmetics).config

    def _quantized_inference(self, x: int) -> int:
        fp_input = self._config.as_rational(x)
        with torch.no_grad():
            output = self(torch.tensor(fp_input))
        return self._config.as_integer(float(output.item()))

    def translate(self, name: str) -> Design:
        quantized_inputs = list(map(self._config.as_integer, self._step_lut.tolist()))
        return PrecomputedMonotonouslyIncreasingScalarFunction(
            name=name,
            width=self._config.total_bits,
            inputs=quantized_inputs,
            function=self._quantized_inference,
        )



def to_unsigned(value: int, total_bits: int) -> int:
    def invert(value: int) -> int:
        return value ^ int("1" * total_bits, 2)

    def discard_leading_bits(value: int) -> int:
        return value & int("1" * total_bits, 2)

    if value < 0:
        value = discard_leading_bits(invert(abs(value)) + 1)

    return value



import torch

from elasticai.creator.base_modules.arithmetics import Arithmetics
from elasticai.creator.base_modules.autograd_functions.binary_quantization import (
    Binarize,
)


class BinaryArithmetics(Arithmetics):
    def quantize(self, a: torch.Tensor) -> torch.Tensor:
        return Binarize.apply(a)

    def clamp(self, a: torch.Tensor) -> torch.Tensor:
        return self.quantize(a)

    def round(self, a: torch.Tensor) -> torch.Tensor:
        return self.quantize(a)

    def add(self, a: torch.Tensor, b: torch.Tensor) -> torch.Tensor:
        return self.quantize(a + b)

    def sum(self, a: torch.Tensor, *others: torch.Tensor) -> torch.Tensor:
        return self.quantize(torch.sum(a, *others))

    def mul(self, a: torch.Tensor, b: torch.Tensor) -> torch.Tensor:
        return self.quantize(a * b)

    def matmul(self, a: torch.Tensor, b: torch.Tensor) -> torch.Tensor:
        return self.quantize(torch.matmul(a, b))



"""
This module provides a parser for manifest.toml manifest files.

this is an example of such a manifest file:

```toml
[elasticai.creator]
version = "==0.34"

```

With the `version` you need to specify which versions of the elastic-ai creator your
`layer.py` is compatible with.
"""

import tomllib
from dataclasses import dataclass


def parse(text: str) -> "Manifest":
    data = tomllib.loads(text)
    data = data["elasticai"]["creator"]
    return Manifest(version=data["version"], layer=data["layer"])


@dataclass
class Manifest:
    version: str
    layer: dict



from functools import partial
from typing import Callable

from elasticai.creator.hdl.code_generation.template import (
    InProjectTemplate,
    module_to_package,
)
from elasticai.creator.hdl.design_base.design import Design, Port
from elasticai.creator.hdl.design_base.signal import Signal
from elasticai.creator.hdl.savable import Path


class PrecomputedMonotonouslyIncreasingScalarFunction(Design):
    _template_package = module_to_package(__name__)

    def __init__(
        self,
        name: str,
        width: int,
        function: Callable[[int], int],
        inputs: list[int],
    ) -> None:
        super().__init__(name)
        self._width = width
        self._function = function
        self._inputs = inputs
        self._template = InProjectTemplate(
            file_name="precomputed_monotonously_increasing_scalar_function.tpl.vhd",
            package=self._template_package,
            parameters=dict(name=self.name, data_width=str(width)),
        )

    def _compute_io_pairs(self) -> dict[int, int]:
        inputs_in_descending_order = sorted(self._inputs, reverse=True)
        pairs = dict()
        for number in inputs_in_descending_order:
            pairs[number] = self._function(number)
        return pairs

    @property
    def port(self) -> Port:
        signal = partial(Signal, width=self._width)
        return Port(
            incoming=[signal(name="x")],
            outgoing=[signal(name="y")],
        )

    def save_to(self, destination: Path) -> None:
        process_content = []

        pairs = list(self._compute_io_pairs().items())
        for input, output in pairs[:-1]:
            process_content.append(
                f"if signed_x <= {input} then signed_y <= to_signed({output},"
                f" {self._width});"
            )
        _, output = pairs[-1]
        process_content.append(f"else signed_y <= to_signed({output}, {self._width});")
        process_content.append("end if;")

        self._template.parameters.update(process_content=process_content)
        destination.create_subpath(self.name).as_file(".vhd").write(self._template)



from collections.abc import Iterator
from typing import cast

import torch

from elasticai.creator.hdl.design_base.design import Design
from elasticai.creator.hdl.translatable import Translatable
from elasticai.creator.hdl.vhdl.designs.sequential import (
    Sequential as _SequentialDesign,
)


class Sequential(Translatable, torch.nn.Sequential):
    def __init__(self, *submodules: Translatable):
        super().__init__(*cast(tuple[torch.nn.Module, ...], submodules))

    def translate(self, name: str) -> Design:
        registry = _Registry()
        submodules: list[Translatable] = [
            cast(Translatable, m) for m in self.children()
        ]
        for module in submodules:
            registry.register(module.__class__.__name__.lower(), module)
        subdesigns = list(registry.build_designs())
        x_address_width = 1
        y_address_width = 1
        if len(subdesigns) == 0:
            x_width = 1
            y_width = 1

        else:
            front = subdesigns[0]
            back = subdesigns[-1]
            x_width = front.port["x"].width
            y_width = back.port["y"].width
            found_y_address = False
            found_x_address = False
            for design in subdesigns:
                if "y_address" in design.port and not found_y_address:
                    found_y_address = True
                    y_address_width = design.port["y_address"].width
                if "x_address" in design.port and not found_x_address:
                    found_x_address = True
                    x_address_width = back.port["x_address"].width
        return _SequentialDesign(
            sub_designs=subdesigns,
            x_width=x_width,
            y_width=y_width,
            x_address_width=x_address_width,
            y_address_width=y_address_width,
            name=name,
        )


class _Registry:
    def __init__(self) -> None:
        self._nodes: dict[str, Translatable] = {}
        self._name_counters: dict[str, int] = {}

    def _make_name_unique(self, name: str) -> str:
        return f"{name}_{self._get_counter_for_name(name)}"

    def _get_counter_for_name(self, name: str) -> int:
        if name in self._name_counters:
            return self._name_counters[name]
        else:
            return 0

    def _increment_name_counter(self, name: str):
        self._name_counters[name] = 1 + self._get_counter_for_name(name)

    def register(self, name: str, d: Translatable):
        unique_name = self._make_name_unique(name)
        self._nodes[unique_name] = d
        self._increment_name_counter(name)

    def build_designs(self) -> Iterator[Design]:
        for name, module in self._nodes.items():
            yield module.translate(name)



""" This file corresponds to the functional module in pytorch. We keep here custom autograd functions
"""

from typing import Any

import torch


class Binarize(torch.autograd.Function):
    @staticmethod
    def forward(ctx: Any, *args: Any, **kwargs: Any) -> Any:
        if len(args) == 0:
            raise TypeError
        x = args[0]
        out_of_range = torch.logical_or(torch.gt(x, 1.0), torch.lt(x, -1.0))
        ctx.save_for_backward(out_of_range)
        y = torch.where(x >= 0, 1.0, -1.0)
        return y

    @staticmethod
    def backward(ctx: Any, *grad_outputs: Any) -> Any:
        (out_of_range,) = ctx.saved_tensors
        return grad_outputs[0] * torch.where(out_of_range, 0.0, 1.0)



from typing import Any, cast

from elasticai.creator.base_modules.linear import Linear
from elasticai.creator.hdl.design_base.design import Design
from elasticai.creator.hdl.translatable import Translatable
from elasticai.creator.nn.fixed_point_arithmetics import FixedPointArithmetics
from elasticai.creator.nn.two_complement_fixed_point_config import FixedPointConfig

from .design import FPLinear as FPLinearDesign


class FPLinear(Translatable, Linear):
    def __init__(
        self,
        in_features: int,
        out_features: int,
        total_bits: int,
        frac_bits: int,
        bias: bool,
        device: Any = None,
    ) -> None:
        self._config = FixedPointConfig(total_bits=total_bits, frac_bits=frac_bits)
        super().__init__(
            in_features=in_features,
            out_features=out_features,
            arithmetics=FixedPointArithmetics(config=self._config),
            bias=bias,
            device=device,
        )

    def translate(self, name: str) -> Design:
        def float_to_signed_int(value: float | list) -> int | list:
            if isinstance(value, list):
                return list(map(float_to_signed_int, value))
            return self._config.as_integer(value)

        bias = [0] * self.out_features if self.bias is None else self.bias.tolist()
        signed_int_weights = cast(
            list[list[int]], float_to_signed_int(self.weight.tolist())
        )
        signed_int_bias = cast(list[int], float_to_signed_int(bias))

        return FPLinearDesign(
            frac_bits=self._config.frac_bits,
            total_bits=self._config.total_bits,
            in_feature_num=self.in_features,
            out_feature_num=self.out_features,
            weights=signed_int_weights,
            bias=signed_int_bias,
            name=name,
        )



from abc import ABC, abstractmethod

from elasticai.creator.hdl.design_base.design import Design


class Translatable(ABC):
    @abstractmethod
    def translate(self, name: str) -> Design:
        ...



from typing import Optional, Protocol

import torch

from elasticai.creator.base_modules.lstm_cell import LSTMCell


class LayerFactory(Protocol):
    def lstm(self, input_size: int, hidden_size: int, bias: bool) -> LSTMCell:
        ...


class LSTM(torch.nn.Module):
    def __init__(
        self,
        input_size: int,
        hidden_size,
        bias: bool,
        batch_first: bool,
        layers: LayerFactory,
    ) -> None:
        super().__init__()
        self.cell = layers.lstm(
            input_size=input_size, hidden_size=hidden_size, bias=bias
        )
        self.batch_first = batch_first

    @property
    def hidden_size(self) -> int:
        return self.cell.hidden_size

    @property
    def input_size(self) -> int:
        return self.cell.input_size

    def forward(
        self,
        x: torch.Tensor,
        state: Optional[tuple[torch.Tensor, torch.Tensor]] = None,
    ) -> tuple[torch.Tensor, tuple[torch.Tensor, torch.Tensor]]:
        batched = x.dim() == 3

        if batched and self.batch_first:
            x = torch.stack(torch.unbind(x), dim=1)

        if state is not None:
            state = state[0].squeeze(0), state[1].squeeze(0)

        inputs = torch.unbind(x, dim=0)

        outputs = []
        for i in range(len(inputs)):
            hidden_state, cell_state = self.cell(inputs[i], state)
            state = (hidden_state, cell_state)
            outputs.append(hidden_state)

        if state is None:
            raise RuntimeError("Number of samples must be larger than 0.")

        result = torch.stack(outputs, dim=1 if batched and self.batch_first else 0)
        # TODO: check whether unsqueeze dimension is actually consistent with self.batch_first being true or false
        hidden_state, cell_state = state[0].unsqueeze(0), state[1].unsqueeze(0)
        return result, (hidden_state, cell_state)



from typing import Callable, Optional

import torch
from torch.nn.utils.parametrize import register_parametrization

from elasticai.creator.base_modules.autograd_functions.binary_quantization import (
    Binarize as BinarizeFn,
)
from elasticai.creator.mlframework import Module, Tensor

"""Implementation of quantizers and quantized variants of pytorch layers"""


class Binarize(torch.nn.Module):
    """
    Implementation of binarizer with possible bits [-1,1]"""

    def __init__(self):
        super().__init__()

    @staticmethod
    def forward(x):
        return BinarizeFn.apply(x)

    @staticmethod
    def right_inverse(x: Tensor) -> Tensor:
        return x


class Ternarize(torch.nn.Module):
    """
    Implementation of ternarizer with possible bits [-1,0,1]
    Args:
     zero_window_width: all numbers x with |x| <= |zero_window_width| will be rounded to zero, *default* value is
     0.5 which corresponds to natural rounding
    """

    def __init__(
        self,
        zero_window_width=0.5,
        trainable=False,
    ):
        super().__init__()

        factor = torch.tensor(0.5 / zero_window_width)

        # naming only possible per axis and is only a prototype feature --> decline
        # requires_grad flag to false makes parameters non-trainable
        self.widening_factor = torch.nn.Parameter(factor, requires_grad=trainable)

    def forward(self, x):
        clipped = torch.clamp(self.widening_factor * x, min=-1, max=1)
        rounded = torch.round(clipped)
        with torch.no_grad():
            no_gradient = rounded - clipped
        return clipped + no_gradient

    @staticmethod
    def right_inverse(x: Tensor) -> Tensor:
        return x


class Identity(torch.nn.Module):
    def __init__(self) -> None:
        super().__init__()

    @staticmethod
    def forward(x: torch.Tensor) -> torch.Tensor:
        return x


class ResidualQuantization(torch.nn.Module):
    """
    Implementation of residual quantization similar to Rebnet: https://arxiv.org/pdf/1711.01243.pdf
    But instead of summing the bits, they are concatenated, multiplying the second dimension by n bits.
    Here the encoded form is used to facilitate precalculation.
    Args:
     num_bits: number of residual quantization levels or bits
    """

    def __init__(self, num_bits=2):
        super().__init__()
        self.num_bits = num_bits
        self.weight = torch.nn.Parameter(
            torch.Tensor([0.5**i for i in range(1, num_bits)])
        )

    def forward(self, inputs):
        bin = BinarizeFn.apply
        out = [inputs]
        for factor in self.weight:
            out.append(out[-1] - torch.abs(factor) * bin(out[-1]))
        return bin(torch.cat(out, dim=1))


class QuantizeTwoBit(torch.nn.Module):
    """
    Implementation of 2 bit residual quantization, the first bit is binarized the second uses  a factor similar to residual quantization
    Args:
      factors: the initial factor
    """

    def __init__(self, factors=0.5):
        super().__init__()
        self.factors = torch.nn.Parameter(torch.Tensor([factors]), requires_grad=True)

    def forward(self, input):
        binarize = BinarizeFn.apply
        first_half = binarize(input)
        second_half = input - self.factors * binarize(input)
        return binarize(torch.cat((first_half, second_half), dim=1))


def _hook_quantizer(module: torch.nn.Module, quantizer: Module, bias: bool):
    if isinstance(quantizer, Module):
        register_parametrization(module, "weight", quantizer)
        if bias:
            register_parametrization(module, "bias", quantizer)
    else:
        raise TypeError(f"Quantizer {quantizer} is not an instance of Module.")


class QConv1d(torch.nn.Conv1d):
    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        kernel_size: int | tuple[int],
        quantizer: Module,
        stride: tuple[int] = (1,),
        padding: int = 0,
        dilation: tuple[int] = (1,),
        groups: int = 1,
        bias: bool = True,
        padding_mode: str = "zeros",
    ):
        super().__init__(
            in_channels=in_channels,
            out_channels=out_channels,
            kernel_size=kernel_size,
            stride=stride,
            padding=padding,
            dilation=dilation,
            groups=groups,
            bias=bias,
            padding_mode=padding_mode,
        )

        _hook_quantizer(self, quantizer=quantizer, bias=bias)


class QConv2d(torch.nn.Conv2d):
    def __init__(
        self,
        in_channels,
        out_channels,
        kernel_size,
        quantizer,
        stride=1,
        padding=0,
        dilation=1,
        groups=1,
        bias=True,
        padding_mode="zeros",
    ):
        super().__init__(
            in_channels=in_channels,
            out_channels=out_channels,
            kernel_size=kernel_size,
            stride=stride,
            padding=padding,
            dilation=dilation,
            groups=groups,
            bias=bias,
            padding_mode=padding_mode,
        )
        _hook_quantizer(self, quantizer=quantizer, bias=bias)


class ChannelShuffle(torch.nn.Module):
    def __init__(self, groups):
        super(ChannelShuffle, self).__init__()
        self.groups = groups

    def forward(self, x):
        num_channels = x.data.size()[1]
        original_shape = x.data.size()[2:]
        x = x.view(
            x.data.size()[0], self.groups, num_channels // self.groups, *original_shape
        )
        x = x.transpose(1, 2).contiguous()
        return x.view(x.data.size()[0], num_channels, *original_shape)


class QLinear(torch.nn.Linear):
    def __init__(
        self,
        in_features: int,
        out_features: int,
        quantizer: Module,
        bias: bool = True,
    ):
        super().__init__(in_features=in_features, out_features=out_features, bias=bias)
        _hook_quantizer(self, quantizer=quantizer, bias=bias)


class _QLSTMCellBase(torch.nn.LSTMCell):
    def __init__(
        self,
        input_size: int,
        hidden_size: int,
        state_quantizer: Module,
        weight_quantizer: Module,
        bias: bool,
        input_gate_activation: Callable[[torch.Tensor], torch.Tensor],
        forget_gate_activation: Callable[[torch.Tensor], torch.Tensor],
        cell_gate_activation: Callable[[torch.Tensor], torch.Tensor],
        output_gate_activation: Callable[[torch.Tensor], torch.Tensor],
        new_cell_state_activation: Callable[[torch.Tensor], torch.Tensor],
    ):
        super().__init__(input_size, hidden_size, bias)

        self.state_quantizer = state_quantizer
        self.weight_quantizer = weight_quantizer
        self.input_gate_activation = input_gate_activation
        self.forget_gate_activation = forget_gate_activation
        self.cell_gate_activation = cell_gate_activation
        self.output_gate_activation = output_gate_activation
        self.new_cell_state_activation = new_cell_state_activation

    def forward(
        self, x: torch.Tensor, state: Optional[tuple[torch.Tensor, torch.Tensor]] = None
    ) -> tuple[torch.Tensor, torch.Tensor]:
        batched = x.dim() == 2
        if state is None:
            zeros_shape = (
                (x.size(dim=0), self.hidden_size) if batched else (self.hidden_size,)
            )
            zeros = torch.zeros(*zeros_shape, dtype=x.dtype, device=x.device)
            state = (zeros, zeros)

        h_0 = self.state_quantizer(state[0])
        c_0 = self.state_quantizer(state[1])
        weight_ih = self.weight_quantizer(self.weight_ih)
        weight_hh = self.weight_quantizer(self.weight_hh)

        gates = torch.matmul(x, weight_ih.t()) + torch.matmul(h_0, weight_hh.t())

        if self.bias:
            bias_ih = self.weight_quantizer(self.bias_ih)
            bias_hh = self.weight_quantizer(self.bias_hh)
            gates += bias_ih + bias_hh

        in_gate, forget_gate, cell_gate, out_gate = gates.chunk(
            4, dim=1 if batched else 0
        )

        i = self.input_gate_activation(in_gate)
        f = self.forget_gate_activation(forget_gate)
        g = self.cell_gate_activation(cell_gate)
        o = self.output_gate_activation(out_gate)

        c_1 = f * c_0 + i * g
        h_1 = o * self.new_cell_state_activation(c_1)

        return h_1, c_1


class _QLSTMBase(torch.nn.Module):
    def __init__(self, lstm_cell: _QLSTMCellBase, batch_first: bool):
        super().__init__()
        self.cell = lstm_cell
        self.batch_first = batch_first

    def forward(
        self,
        x: torch.Tensor,
        state: Optional[tuple[torch.Tensor, torch.Tensor]] = None,
    ) -> tuple[torch.Tensor, tuple[torch.Tensor, torch.Tensor]]:
        batched = x.dim() == 3

        if batched and self.batch_first:
            x = torch.stack(torch.unbind(x), dim=1)

        if state is not None:
            state = state[0].squeeze(0), state[1].squeeze(0)

        inputs = torch.unbind(x, dim=0)

        outputs = []
        for i in range(len(inputs)):
            hidden_state, cell_state = self.cell(inputs[i], state)
            state = (hidden_state, cell_state)
            outputs.append(hidden_state)

        result = torch.stack(outputs, dim=1 if batched and self.batch_first else 0)
        hidden_state, cell_state = state[0].unsqueeze(0), state[1].unsqueeze(0)

        return result, (hidden_state, cell_state)


class QLSTMCell(_QLSTMCellBase):
    def __init__(
        self,
        input_size: int,
        hidden_size: int,
        state_quantizer: Module,
        weight_quantizer: Module,
        bias: bool = True,
    ) -> None:
        super().__init__(
            input_size=input_size,
            hidden_size=hidden_size,
            state_quantizer=state_quantizer,
            weight_quantizer=weight_quantizer,
            bias=bias,
            input_gate_activation=torch.sigmoid,
            forget_gate_activation=torch.sigmoid,
            cell_gate_activation=torch.tanh,
            output_gate_activation=torch.sigmoid,
            new_cell_state_activation=torch.tanh,
        )


class QLSTM(_QLSTMBase):
    def __init__(
        self,
        input_size: int,
        hidden_size: int,
        state_quantizer: Module,
        weight_quantizer: Module,
        bias: bool = True,
        batch_first: bool = False,
    ) -> None:
        super().__init__(
            lstm_cell=QLSTMCell(
                input_size=input_size,
                hidden_size=hidden_size,
                state_quantizer=state_quantizer,
                weight_quantizer=weight_quantizer,
                bias=bias,
            ),
            batch_first=batch_first,
        )



from elasticai.creator.hdl.code_generation.template import (
    InProjectTemplate,
    TemplateExpander,
    module_to_package,
)

"""
We want to programmatically use designs that are either hand-written or based on hand-written templates
these designs need to adhere to a well defined protocol. As a hardware designer you specify your protocol
and the expected version of the elasticai.creator in a file called `design_meta.toml` that lives in the
same folder as your `layer.py` that defines how the hdl code for the hardware design and the behaviour of
the corresponding neural network layer in software.

To help you as hardware designer stick to specific protocol you can generate a base template, that you
can use as a starting point to develop your design.
"""


class BaseTemplateGenerator:
    def __init__(self) -> None:
        self._base_template = InProjectTemplate(
            package=module_to_package(self.__module__),
            file_name="base_template.tpl.vhd",
            parameters={},
        )

    def generate(self) -> str:
        return self._expand_base_template()

    def _expand_base_template(self) -> str:
        return "\n".join(TemplateExpander(self._base_template).lines())



from itertools import chain

from elasticai.creator.hdl.code_generation.template import (
    InProjectTemplate,
    module_to_package,
)
from elasticai.creator.hdl.design_base.design import Design, Port
from elasticai.creator.hdl.design_base.std_signals import (
    clock,
    done,
    enable,
    x,
    x_address,
    y,
    y_address,
)
from elasticai.creator.hdl.savable import Path
from elasticai.creator.hdl.vhdl.code_generation import create_instance
from elasticai.creator.hdl.vhdl.code_generation.code_generation import (
    create_connections_using_to_from_pairs,
    create_signal_definitions,
)


class Sequential(Design):
    def __init__(
        self,
        sub_designs: list[Design],
        *,
        x_width: int,
        y_width: int,
        x_address_width: int,
        y_address_width: int,
        name: str,
    ) -> None:
        super().__init__(name)
        self._x_width = x_width
        self._y_width = y_width
        self._x_address_width = x_address_width
        self._y_address_width = y_address_width
        self._library_name_for_instances = "work"
        self._architecture_name_for_instances = "rtl"
        self._subdesigns = sub_designs

    def _qualified_signal_name(self, instance: str, signal: str) -> str:
        return f"{instance}_{signal}"

    @property
    def port(self) -> Port:
        return Port(
            incoming=[
                x(width=self._x_width),
                y_address(width=self._y_address_width),
                clock(),
                enable(),
            ],
            outgoing=[
                y(width=self._y_width),
                x_address(width=self._x_address_width),
                done(),
            ],
        )

    def _save_subdesigns(self, destination: Path) -> None:
        for design in self._subdesigns:
            design.save_to(destination.create_subpath(design.name))

    def _instance_names(self) -> list[str]:
        return [f"i_{design.name}" for design in self._subdesigns]

    def _generate_connections(self) -> list[str]:
        prefixes = [f"{name}_" for name in self._instance_names()]
        connections = {f"{name}clock": "clock" for name in prefixes}

        def wire_pair(a, b):
            connections.update(
                {
                    f"{a}y_address": f"{b}x_address",
                    f"{b}x": f"{a}y",
                    f"{b}enable": f"{a}done",
                }
            )

        for a, b in zip(prefixes[:-1], prefixes[1:]):
            wire_pair(a, b)

        def wire_top_to_start_and_end(start, end):
            connections.update(
                {
                    "done": f"{end}done",
                    "y": f"{end}y",
                    f"{end}y_address": f"y_address",
                    f"{start}x": "x",
                    f"{start}enable": "enable",
                    f"x_address": f"{start}x_address",
                }
            )

        if len(prefixes) == 0:
            connections.update({"x_address": "y_address", "y": "x", "done": "enable"})
        else:
            wire_top_to_start_and_end(prefixes[0], prefixes[-1])

        return create_connections_using_to_from_pairs(connections)

    def _instance_name_and_design_pairs(self):
        yield from zip(self._instance_names(), self._subdesigns)

    def _generate_instantiations(self) -> list[str]:
        instantiations: list[str] = list()
        for instance, design in self._instance_name_and_design_pairs():
            signal_map = {
                signal.name: self._qualified_signal_name(instance, signal.name)
                for signal in design.port
            }
            instantiations.extend(
                create_instance(
                    name=instance,
                    entity=design.name,
                    library=self._library_name_for_instances,
                    architecture=self._architecture_name_for_instances,
                    signal_mapping=signal_map,
                )
            )
        return instantiations

    def _generate_signal_definitions(self) -> list[str]:
        return sorted(
            chain.from_iterable(
                create_signal_definitions(f"{instance_id}_", instance.port.signals)
                for instance_id, instance in self._instance_name_and_design_pairs()
            )
        )

    def save_to(self, destination: Path):
        self._save_subdesigns(destination)
        network_template = InProjectTemplate(
            package=module_to_package(self.__module__),
            file_name="network.tpl.vhd",
            parameters=dict(
                layer_connections=self._generate_connections(),
                layer_instantiations=self._generate_instantiations(),
                signal_definitions=self._generate_signal_definitions(),
                x_address_width=str(self._x_address_width),
                y_address_width=str(self._y_address_width),
                x_width=str(self._x_width),
                y_width=str(self._y_width),
                layer_name=self.name,
            ),
        )
        destination.create_subpath(self.name).as_file(".vhd").write(network_template)



from elasticai.creator.hdl.code_generation.code_generation import (
    calculate_address_width,
)
from elasticai.creator.hdl.code_generation.template import (
    InProjectTemplate,
    module_to_package,
)
from elasticai.creator.hdl.design_base.design import Design, Port
from elasticai.creator.hdl.design_base.ports import create_port_for_buffered_design
from elasticai.creator.hdl.savable import Path


class Identity(Design):
    def __init__(self, name: str, num_input_features: int, num_input_bits: int) -> None:
        self._num_input_features = num_input_features
        self._num_input_bits = num_input_bits
        self._address_width = calculate_address_width(num_input_features)
        super().__init__(name)

    @property
    def port(self) -> Port:
        return create_port_for_buffered_design(
            x_width=self._num_input_bits,
            y_width=self._num_input_bits,
            x_count=self._num_input_features,
            y_count=self._num_input_features,
        )

    def save_to(self, destination: Path) -> None:
        template = InProjectTemplate(
            package=module_to_package(self.__module__),
            file_name="identity.tpl.vhd",
            parameters=dict(
                name=self.name,
                x_address_width=str(self._address_width),
                y_address_width=str(self._address_width),
                x_width=str(self._num_input_bits),
                y_width=str(self._num_input_bits),
            ),
        )
        destination.create_subpath(self.name).as_file(".vhd").write(template)



from abc import ABC

from elasticai.creator.hdl.code_generation.code_generation import (
    calculate_address_width,
)
from elasticai.creator.hdl.design_base import std_signals as _signals
from elasticai.creator.hdl.design_base.design import Design, Port


class NetworkBlock(Design, ABC):
    def __init__(self, name: str, x_width: int, y_width: int):
        super().__init__(name)
        self._port = Port(
            incoming=[
                _signals.enable(),
                _signals.clock(),
                _signals.x(x_width),
            ],
            outgoing=[_signals.y(y_width)],
        )


class BufferedNetworkBlock(Design, ABC):
    def __init__(
        self,
        name: str,
        x_width: int,
        y_width: int,
        x_count: int,
        y_count: int,
    ):
        super().__init__(name)
        in_signals = [
            _signals.enable(),
            _signals.clock(),
            _signals.x(x_width),
            _signals.y_address(calculate_address_width(y_count)),
        ]
        out_signals = [
            _signals.done(),
            _signals.y(y_width),
            _signals.x_address(calculate_address_width(x_count)),
        ]
        self._port = Port(incoming=in_signals, outgoing=out_signals)

    @property
    def port(self) -> Port:
        return self._port



import torch

from elasticai.creator.base_modules.arithmetics import Arithmetics


class FloatArithmetics(Arithmetics):
    def quantize(self, a: torch.Tensor) -> torch.Tensor:
        return a

    def clamp(self, a: torch.Tensor) -> torch.Tensor:
        return a

    def round(self, a: torch.Tensor) -> torch.Tensor:
        return a

    def add(self, a: torch.Tensor, b: torch.Tensor) -> torch.Tensor:
        return a + b

    def sum(self, tensor: torch.Tensor, *tensors: torch.Tensor) -> torch.Tensor:
        summed = tensor
        for t in tensors:
            summed += t
        return summed

    def mul(self, a: torch.Tensor, b: torch.Tensor) -> torch.Tensor:
        return a * b

    def matmul(self, a: torch.Tensor, b: torch.Tensor) -> torch.Tensor:
        return torch.matmul(a, b)



"""Interfaces/Adapters to the used machine learning framework"""
from abc import abstractmethod
from typing import (
    Any,
    Callable,
    Iterable,
    Iterator,
    Protocol,
    TypeVar,
    runtime_checkable,
)

from numpy.typing import ArrayLike

Index = int | slice | tuple[int | slice]


class Tensor(Protocol):
    def __getitem__(self, index: Index) -> "Tensor":
        ...

    def apply_(self, callable: Callable) -> "Tensor":
        ...

    def detach(self) -> "Tensor":
        ...

    def numpy(self) -> ArrayLike:
        ...


@runtime_checkable
class Parameter(Tensor, Protocol):
    ...


T_Module = TypeVar("T_Module", bound="Module")


@runtime_checkable
class Module(Protocol):
    @abstractmethod
    def train(self: T_Module, mode: bool = True) -> T_Module:
        ...

    @abstractmethod
    def extra_repr(self) -> str:
        ...

    @abstractmethod
    def named_children(self) -> Iterable[tuple[str, "Module"]]:
        ...

    @abstractmethod
    def parameters(self, recurse: bool = True) -> Iterator[Parameter]:
        ...

    @abstractmethod
    def state_dict(self) -> dict[str, Any]:
        ...

    @abstractmethod
    def __call__(self, x: Any, *args: Any, **kwargs: Any) -> Any:
        ...



from typing import Protocol

import torch


class Arithmetics(Protocol):
    def quantize(self, a: torch.Tensor) -> torch.Tensor:
        ...

    def clamp(self, a: torch.Tensor) -> torch.Tensor:
        ...

    def round(self, a: torch.Tensor) -> torch.Tensor:
        ...

    def add(self, a: torch.Tensor, b: torch.Tensor) -> torch.Tensor:
        ...

    def sum(self, tensor: torch.Tensor, *tensors: torch.Tensor) -> torch.Tensor:
        ...

    def mul(self, a: torch.Tensor, b: torch.Tensor) -> torch.Tensor:
        ...

    def matmul(self, a: torch.Tensor, b: torch.Tensor) -> torch.Tensor:
        ...



import math


def calculate_address_width(num_items: int) -> int:
    return max(1, math.ceil(math.log2(num_items)))


def to_hex(number: int, bit_width: int) -> str:
    return f"{number:0{math.ceil(bit_width / 4)}x}"



from typing import cast

import torch

from elasticai.creator.base_modules.arithmetics import Arithmetics
from elasticai.creator.base_modules.autograd_functions.identity_step_function import (
    IdentityStepFunction,
)


class Tanh(torch.nn.Tanh):
    def __init__(
        self,
        arithmetics: Arithmetics,
        num_steps: int,
        sampling_intervall: tuple[float, float],
    ) -> None:
        super().__init__()
        self._arithmetics = arithmetics
        self._step_lut = torch.linspace(*sampling_intervall, num_steps)

    def _quantized_step_inputs(self, inputs: torch.Tensor) -> torch.Tensor:
        step_inputs = cast(
            torch.Tensor, IdentityStepFunction.apply(inputs, self._step_lut)
        )
        return self._arithmetics.quantize(step_inputs)

    def forward(self, inputs: torch.Tensor) -> torch.Tensor:
        inputs = self._quantized_step_inputs(inputs)
        outputs = torch.nn.functional.tanh(inputs)
        return self._arithmetics.quantize(outputs)



import re
from abc import abstractmethod
from typing import Protocol, Sequence, overload

from elasticai.creator.hdl.code_generation.code_generation import to_hex


def _sorted_dict(items: dict[str, str]) -> dict[str, str]:
    return dict((key, items[key]) for key in sorted(items))


def create_instance(
    *,
    name: str,
    entity: str,
    signal_mapping: dict[str, str],
    library: str,
    architecture: str = "rtl",
) -> list[str]:
    signal_mapping = _sorted_dict(signal_mapping)
    result = [f"{name} : entity {library}.{entity}({architecture})", "port map("]
    for _from in tuple(signal_mapping.keys())[:-1]:
        _to = signal_mapping[_from]
        result.append(f"  {_from} => {_to},")
    last_from = tuple(signal_mapping.keys())[-1]
    last_to = signal_mapping[last_from]
    result.append(f"  {last_from} => {last_to}")
    result.append(");")
    return result


def create_connections_using_to_from_pairs(mapping: dict[str, str]) -> list[str]:
    mapping = _sorted_dict(mapping)
    connections: list[str] = []
    for _to, _from in mapping.items():
        connections.append(f"{_to} <= {_from};")
    return connections


class Signal(Protocol):
    @property
    @abstractmethod
    def name(self) -> str:
        ...

    @property
    @abstractmethod
    def width(self) -> int:
        ...


def create_signal_definitions(prefix: str, signals: Sequence[Signal]):
    return sorted(
        [
            signal_definition(name=f"{prefix}{signal.name}", width=signal.width)
            for signal in signals
        ]
    )


def signal_definition(
    *,
    name: str,
    width: int,
):
    def vector_signal(name: str, width) -> str:
        return (
            f"signal {name} : std_logic_vector({width - 1} downto 0)"
            " := (others => '0');"
        )

    def logic_signal(name: str) -> str:
        return f"signal {name} : std_logic := '0';"

    if width > 0:
        return vector_signal(name, width)
    else:
        return logic_signal(name)


def hex_representation(hex_value: str) -> str:
    return f'x"{hex_value}"'


def bin_representation(bin_value: str) -> str:
    return f'"{bin_value}"'


def to_vhdl_hex_string(number: int, bit_width: int) -> str:
    return f"'x{to_hex(number, bit_width)}'"


def generate_hex_for_rom(value: str):
    return f'x"{value}"'


@overload
def extract_rom_values(text: str) -> tuple[str]:
    ...


@overload
def extract_rom_values(text: list[str]) -> tuple[str]:
    ...


def extract_rom_values(text: str | list[str]) -> tuple[str, ...]:
    if not isinstance(text, list):
        text = [text]
    values: tuple[str, ...] = tuple()
    for line in text:
        match = re.match(
            r'.*\(x"([a-f0-9]+(",\s?x"[a-f0-9]+)*)"\)',
            line,
        )
        if match is not None:
            array = match.group(1)
            values = tuple(re.split(r'(?:",\s?x")', array))

    return values



from functools import partial

from elasticai.creator.hdl.code_generation.code_generation import (
    calculate_address_width,
)
from elasticai.creator.hdl.code_generation.template import (
    InProjectTemplate,
    module_to_package,
)
from elasticai.creator.hdl.savable import Path
from elasticai.creator.hdl.vhdl.code_generation.twos_complement import to_unsigned


class Rom:
    def __init__(
        self, name: str, data_width: int, values_as_integers: list[int]
    ) -> None:
        self._name = name
        self._data_width = data_width
        number_of_values = len(values_as_integers)
        self._address_width = self._bits_required_to_address_n_values(number_of_values)
        self._values = self._append_zeros_to_fill_addressable_memory(
            self._values_to_unsigned_integers(values_as_integers)
        )

    def save_to(self, destination: Path):
        template = InProjectTemplate(
            file_name="rom.tpl.vhd",
            package=module_to_package(self.__module__),
            parameters=dict(
                rom_value=self._rom_values(),
                rom_addr_bitwidth=str(self._address_width),
                rom_data_bitwidth=str(self._data_width),
                name=self._name,
                resource_option="auto",
            ),
        )
        destination.as_file(".vhd").write(template)

    def _values_to_unsigned_integers(self, values: list[int]) -> list[int]:
        to_uint = partial(to_unsigned, total_bits=self._data_width)
        return list(map(to_uint, values))

    def _rom_values(self) -> str:
        values = [self._format_string_for_rom_values().format(x) for x in self._values]
        return ",".join(values)

    def _format_string_for_rom_values(self) -> str:
        return 'x"{{:0>{num_of_nibbles}x}}"'.format(
            num_of_nibbles=self._number_of_nibbles()
        )

    def _append_zeros_to_fill_addressable_memory(self, values: list[int]) -> list[int]:
        missing_number_of_zeros = 2**self._address_width - len(values)
        return values + self._zeros(missing_number_of_zeros)

    def _bits_required_to_address_n_values(self, n: int) -> int:
        return calculate_address_width(n)

    @staticmethod
    def _zeros(n: int) -> list[int]:
        return [0] * n

    def _number_of_nibbles(self) -> int:
        nibble_length = 4
        return self._data_width // nibble_length
